require(deSolve)
require(tidyverse)
require(magrittr)

season_beta <- function(time, beta, epsilon) beta * (1 + epsilon * sin(pi * time)^6)

# these are the definitions of the derivatives
dzdt_multipop <- function(z, p, lambda) (1 - z) * lambda - p$mu * z + t(p$chi) %*% z
dwdt_multipop <- function(w, p, sum_shared_lambda) (1 - w) * sum_shared_lambda - p$mu * w + t(p$chi) %*% w
dydt_multipop <- function(y, z, w, p, lambda) {
  ((1 - w) + (1 - p$gamma) * (w - z)) * lambda - p$sigma * y - p$nu * y + t(p$chi) %*% y
}
all_multipop_derivs <-  function(t, state, params) {
  # if (any(state < 0)) browser() # state <- ifelse(state < 10^-12, 0, state)
  currently_infectious <- matrix(state[str_detect(names(state), "\\.y")],
                                 nrow=length(state) / params$h$n_strains / 3, ncol=params$h$n_strains) %>%
    set_colnames(state %>% names() %>% str_extract("Strain_\\d+") %>% unique()) %>%
    set_rownames(state %>% names() %>% str_extract("Population_\\d+") %>% unique()) %>%
    as.data.frame()
  # do the necessary sums for this timestep
  # total amount of virus in each population
  params$h$totVirus <- currently_infectious %>% rowSums()
  # address seasonality in beta
  betai <- season_beta(t, params$p$beta, params$p$epsilon)
  # force of infection in z equations
  params$h$lambda <- currently_infectious * betai
  # sum of virus that shares epitopes with each strain
  # [i.e. force of infection for each virus in w equation]
  params$h$sum_shared_lambda <- lapply(1:params$h$n_strains, function(strain) {
    rowSums(currently_infectious[,params$h$shared_epitopes[strain,]] * betai[,strain])
  }) %>% set_names(names(params$h$lambda)) %>% do.call(cbind, .) %>% as.data.frame()
  # calculate the derivatives
  list(lapply(names(state) %>% str_extract("\\.\\w+\\.") %>% unique(), function(strain) {
    c(
      dydt_multipop(state[str_detect(names(state), str_c(strain, "y"))],
                    state[str_detect(names(state), str_c(strain, "z"))],
                    state[str_detect(names(state), str_c(strain, "w"))],
                    params$p, params$h$lambda[[strain %>% str_replace_all("\\.", "")]]),
      dzdt_multipop(state[str_detect(names(state), str_c(strain, "z"))],
                    params$p, params$h$lambda[[strain %>% str_replace_all("\\.", "")]]),
      dwdt_multipop(state[str_detect(names(state), str_c(strain, "w"))],
                    params$p, params$h$sum_shared_lambda[[strain %>% str_replace_all("\\.", "")]])
    )
  }) %>% unlist() %>% set_names(names(state)))
}

run_multipop_mantis <- function(struct, t_max, t_step, init_cond,
                                beta, gamma, sigma, chi, mu=0, nu=0, epsilon=0, delta=NA) {

  n.loci <- length(struct)
  strainmx <- as.data.frame(expand.grid(lapply(struct, seq, from=1, by=1)))
  n_strains <- dim(strainmx)[1]
  colnames(strainmx) <- paste("locus", 1:n.loci, sep="")
  rownames(strainmx) <- NULL
  n_populations <- nrow(chi)

  if (length(beta) == 1) beta <- matrix(beta, n_populations, n_strains)

  for (i in 1:n_strains) {
    temp <- which(strainmx[, 1] == strainmx[i, 1])
    for (j in 2:n.loci) {
      temp <- union(temp, which(strainmx[, j] == strainmx[i, j]))
    }
    if (i == 1) shared_epitopes <- matrix(0, n_strains, length(temp))
    shared_epitopes[i,] <- sort(temp)
  }

  nShEp <- dim(shared_epitopes)[2]

  initial_conditions <- rep(init_cond, each=3) %>%
    multiply_by(c(1,1,nShEp)) %>%
    pmin(rep(1, length(.))) %>%
    set_names(expand.grid(c("y", "z", "w"),
                          str_c("Population_", 1:n_populations),
                          str_c("Strain_", strainmx %>% unite("ss", sep="") %>% use_series(ss))) %>%
                select(Var2, Var3, Var1) %>%
                unite("x", sep=".") %>% use_series(x))

  # cat("Running MANTIS, please wait...\n")
  #assign parameter values to 'p'
  p <- list(beta    = beta,
            gamma   = gamma,
            mu      = mu,
            nu      = nu,
            sigma   = sigma,
            epsilon = epsilon,
            chi     = chi)
  #assign housekeeping values to 'h'
  h <- list(n_strains       = n_strains,
            nShEp           = nShEp,
            shared_epitopes = shared_epitopes)
  #run through the simulation
  ode_out <- ode(y=initial_conditions, times=seq(0, t_max, t_step),
                   func=all_multipop_derivs, parms=list("p"=p, "h"=h))
  return(list(timeseries=ode_out, parameters=p, housekeeping=h, "strains"=list(strainmx)))
}

