library(tidyverse)
library(magrittr)
library(tidygraph)
library(parallel)
library(JuliaCall)
library(ggraph)
library(facetscales)

julia_setup()
julia_source("multipop_MANTIS.jl")

theme_set(theme_bw())

scales_y <- list(
  `currently infectious`    = scale_y_continuous(),
  `specific immunity`       = scale_y_continuous(limits = c(0, 1)),
  `cross-reactive immunity` = scale_y_continuous(limits = c(0, 1))
)

## structural parameters
# set.seed(0)
struct <- c(2, 2)
strains <- expand.grid(lapply(struct, seq, from=1, by=1) %>%
                         setNames(str_c("locus", 1:length(.))))
n_populations <- 1
initial_conditions <- runif(prod(struct) * n_populations) %>%
  matrix(n_populations, prod(struct)) %>%
  apply(1, . %>% {. / (5 * sum(.))}) %>%
  t()

n_populations <- 1
initial_conditions <- matrix(initial_conditions, n_populations, nrow(strains), byrow=TRUE)

## dynamical parameters
gamma <- 0.77 #runif(n_populations, 0.25, 0.95)    # partial cross-protective immunity (cpi)
sigma <- 7.5               # recovery rate
mu <- 0.1                # mortality rate
beta <- 10# * (sigma + mu)               # Infection rate
# delta <- NA               # increase in cpi per allele (not yet implemented)
# epsilon <- 0              # seasonality
# chi <- matrix(0,1,1)
tmax <- 1000

# movement_rate <- 0.05
# movement_network <-
#   tbl_graph(edges=tibble(from=c(1), to=c(2))) %>%
#   # play_erdos_renyi(n_populations, 0.33, directed=TRUE) %>%
#   # play_smallworld(2, sqrt(n_populations), n_populations/8, 0.75) %>% #mutate(x = graph_size() / graph_order()^2)
#   # play_barabasi_albert(n_populations, 2, 15, directed=TRUE) %>%
#   # simple_tree(n_populations, 3, 4, 8) %>%
#   activate(edges) %>%
#   mutate(weight = movement_rate) %>%
#   full_join(activate(., nodes) %>%
#               mutate(indeg = centrality_degree(weights=weight, mode="in"),
#                      outdeg = centrality_degree(weights=weight, mode="out"),
#                      n=1:n()) %>%
#               transmute(from=n, to=n, weight=-outdeg) %>%
#               filter(weight != 0) %>%
#               as_tibble())
#
# chi <- movement_network %>%
#   as_tibble() %>%
#   complete(from=1:n_populations, to=1:n_populations, fill=list("weight"=0)) %>%
#   mutate(from = str_c("Population_", from),
#          to   = str_c("Population_", to)) %>%
#   spread(to, weight) %>%
#   column_to_rownames("from") %>%
#   as.matrix()

chi <- diag(0, n_populations)

## run the simulation
mantis_output <- julia_call("runMANTIS", strainstructure=struct,
                            tstep=(tmax-500):tmax, tmax=tmax,
                            beta=beta, gamma=gamma, sigma=sigma, chi=chi, mu=mu,
                            initvals=initial_conditions)

timeseries <- mantis_output$timeseries %>%
  set_colnames(c(expand.grid(str_c("Population_", 1:n_populations),
                             str_c("Strain_", apply(strains, 1, str_c, collapse="")),
                             c("y", "z", "w")) %>% apply(1, str_c, collapse="."),
                 "time")) %>%
  gather("details", "prevalence", -time) %>%
  mutate(prevalence = round(prevalence, 11)) %>%
  separate(details, c("population", "strain", "equation"), "\\.") %>%
  mutate(variable = factor(equation, levels=c("z", "w", "y"),
                           labels=c("specific immunity",
                                    "cross-reactive immunity",
                                    "currently infectious")))

# plot a timeseries of the dynamics
# ggplot(timeseries %>% filter(time < 20)) +
ggplot(timeseries %>% filter(population == "Population_1")) + # %>% filter(strain == "Strain_11")) +
  aes(colour=population, y=prevalence, x=time) +
  geom_line() +
  facet_grid(variable~strain, scales="free") +
  # facet_grid_sc(rows=vars(variable), cols=vars(population, strain), scales=list(y=scales_y)) +
  scale_color_viridis_d() +
  theme_bw() +
  theme(legend.position="bottom",
        legend.background=element_blank(),
        legend.title=element_blank())

# ggraph(movement_network %>%
#          activate("nodes") %>%
#          mutate(gamma=gamma, name = 1:n()) %>%
#          left_join(timeseries %>%
#                      group_by(population, strain) %>%
#                      filter(time > 0.75 * max(time),
#                             variable == "specific immunity") %>%
#                      summarise(zzz = length(unique(round(prevalence, 5)))) %>%
#                      transmute(dynamics = case_when(zzz == 0 ~ "extinct",
#                                                     zzz == 1 ~ "steady",
#                                                     zzz > 1  ~ "cycles")) %>%
#                      distinct() %>%
#                      mutate(name = str_extract(population, "\\d+") %>%
#                               as.numeric()))) +
#   geom_edge_link(arrow = arrow(length = unit(3, 'mm'), type="closed", angle=15),
#                  start_cap = circle(5, 'mm'),
#                  end_cap = circle(5, 'mm')) +
#   geom_node_point(size = 10, aes(colour = dynamics)) +
#   geom_node_text(aes(label = round(gamma, 2))) +
#   geom_node_label(aes(label = population), nudge_y=-0.25) +
#   theme(axis.text=element_blank(),
#         axis.title=element_blank(),
#         axis.ticks=element_blank())
